import collections
from Parameter import Parameter, SpatialParameter
from IPython.display import display, Latex, HTML
import abc



class SpatialModel(object):
    __metaclass__           = abc.ABCMeta # Make the user code certain
    

    def __init__(self):

        import collections        
        self.setup()





